import os

try:
    import uiautomator2 as u2
except ImportError:
    raise ImportError("Please install 'uiautomator2' library")
from poium.config import App


def connect():
    """
    设备连接
    """
    if App.connect_usb:
        if App.device_id is None:
            device_id = get_device_id()
            App.device_id = device_id

        d = u2.connect_usb(App.device_id)

        return d


def start_app(driver, apk=None):
    """
    启动APP
    """
    if apk is None:
        apk = App.apk_name
    driver.app_start(apk, use_monkey=True)
    driver.app_wait(apk, front=True, timeout=App.app_wait)
    driver.jsonrpc.setConfigurator({"waitForIdleTimeout": 100})

    session = driver.session(apk, attach=True)
    session.implicitly_wait(App.implicitly_wait)
    return session


def close_app(driver, apk=None):
    """
    关闭APP
    """
    driver.app_stop(apk) if apk is not None else driver.app_stop(App.apk_name)


def get_device_id():
    """
    获取设备的 device ID
    """
    status_code = os.system("adb devices")
    if status_code != 0:
        raise SystemError("Verify that adb is properly installed and started")

    _list = os.popen("adb devices | grep -w 'device' | head -1 ")
    device = _list.read()

    if device is None:
        raise NameError("")
    else:
        device_id = device.split("\t")[0]

        return device_id



from selenium.webdriver.common.by import By


def selection_checker(selector: str) -> (str, str):
    """
    check the location method
    :param selector:
    :return:
    """
    if selector.startswith("text=") and len(selector) > 5:
        # link_text
        k = By.LINK_TEXT
        v = selector[5:]
    elif selector.startswith("text~=") and len(selector) > 6:
        # partial_link_text
        k = By.PARTIAL_LINK_TEXT
        v = selector[6:]
    elif selector.startswith("id=") and len(selector) > 3:
        # id
        k = By.ID
        v = selector[3:]
    elif selector.startswith("name=") and len(selector) > 5:
        # name
        k = By.NAME
        v = selector[5:]
    elif selector.startswith("class=") and len(selector) > 6:
        # class name
        k = By.CLASS_NAME
        v = selector[6:]
    elif selector.startswith("tag=") and len(selector) > 4:
        # tag name
        k = By.TAG_NAME
        v = selector[4:]
    elif selector.startswith("/") and len(selector) > 1:
        # xpath
        k = By.XPATH
        v = selector
    else:
        # css
        k = By.CSS_SELECTOR
        v = selector

    return k, v



from poium.base import BaseMethod
from poium.common import logging
from poium.common.assert_des import insert_assert
from poium.config import App

LOCATOR_LIST = [
    "text",
    "textContains",
    "textMatches",
    "textStartsWith",
    "className",
    "classNameMatches",
    "description",
    "descriptionContains",
    "descriptionMatches",
    "descriptionStartsWith",
    "checkable",
    "checked",
    "clickable",
    "longClickable",
    "scrollable",
    "enabled",
    "focusable",
    "focused",
    "selected",
    "packageName",
    "packageNameMatches",
    "resourceId",
    "resourceIdMatches",
    "index",
    "instance",
]


class Page(BaseMethod):
    """
    uiautomator2 page class
    """

    def __init__(self, dr):
        self.driver = dr

    def window_size(self):
        """
        Documents:
            Resolution size

        Return:
            return (width, height)
        """
        return self.driver.window_size()

    def app_list(self):
        """
        Returns:
            list of apps by filter
        """
        return self.driver.app_list()

    def app_info(self, pkg_name=App.apk_name):
        """
        Get app info

        Args:
            pkg_name (str): package name

        Return example:
            {
                "mainActivity": "com.github.uiautomator.MainActivity",
                "label": "ATX",
                "versionName": "1.1.7",
                "versionCode": 1001007,
                "size":1760809
            }

        Raises:
            UiaError
        """
        return self.driver.app_info(pkg_name)

    def click(self, x: float = None, y: float = None, text: str = None, screenshots=App.click_screenshots):
        """
        Args:
            x : width / percentage of width
            y : height / percentage of height
            text: text
            screenshots : screenshots
        Documents:
            click position
        """
        if not x and not y and not text:
            raise ValueError
        (x, y) = self.driver(text=text).center() if text else (x, y)
        self.screenshots(x, y, describe="点击") if screenshots else \
            (print("\n"), logging.info(msg=" 点击 ==> " + "点击坐标{},{}".format(x, y)))

        self.driver.click(x, y)

    def click_more(self, x, y, sleep=.01, times=3):
        """
        连续点击
        x: 横坐标
        y: 纵坐标
        sleep(float): 间隔时间
        times(int): 点击次数
        """
        w, h = self.window_size()
        x, y = (w * x, h * y) if x < 1 and y < 1 else x, y
        for i in range(times):
            self.driver.touch.down(x, y)
            self.sleep(sleep)
            self.driver.touch.up(x, y)

    def swipe(self, fx, fy, tx, ty, duration=0.1, steps=None):
        """
        Args:
            fx: from position
            fy: from position
            tx: to position
            ty: to position
            duration (float): duration
            steps: 1 steps is about 5ms, if set, duration will be ignore

        Documents:
            UiAutoMatTor use steps instead of duration
            As the document say: Each step execution is throttled to 5ms per step.
        """
        self.driver.swipe(fx, fy, tx, ty, duration, steps)

    def swipe_down(self, fx=0.5, fy=0.5, tx=0.5, ty=0.2, duration=0.1, times=1, between=0):
        """
        Args:
            fx: from position
            fy: from position
            tx: to position
            ty: to position
            duration (float): duration
            times (int): Slide number
            between (float): Time interval between
        Documents:
            To the following
        """
        for i in range(times):
            self.swipe(fx, fy, tx, ty, duration=duration, steps=None)
            self.sleep(between)

    def swipe_up(self, fx=0.5, fy=0.5, tx=0.5, ty=0.8, duration=0.1, times=1, between=0):
        """
        Args:
            fx: from position
            fy: from position
            tx: to position
            ty: to position
            duration (float): duration
            times (int): Slide number
            between (float): Time interval between

        Documents:
            To the above
        """
        for i in range(times):
            self.swipe(fx, fy, tx, ty, duration=duration, steps=None)
            self.sleep(between)

    def swipe_left(self, fx=0.3, fy=0.5, tx=0.7, ty=0.5, duration=0.1, times=1, between=0):
        """
        Args:
            fx: from position
            fy: from position
            tx: to position
            ty: to position
            duration (float): duration
            times (int): Slide number
            between (float): Time interval between


        Documents:
            Slide to the left
        """
        for i in range(times):
            self.swipe(fx, fy, tx, ty, duration=duration, steps=None)
            self.sleep(between)

    def swipe_right(self, fx=0.7, fy=0.5, tx=0.3, ty=0.5, duration=0.1, times=1, between=0):
        """
        Args:
            fx: from position
            fy: from position
            tx: to position
            ty: to position
            duration (float): duration
            times (int): Slide number
            between (float): Time interval between

        Documents:
            Slide to the left
        """
        for i in range(times):
            self.swipe(fx, fy, tx, ty, duration=duration, steps=None)
            self.sleep(between)

    def swipe_search(self, text, direction="down", x: float = None, y: float = None):
        """
        文本搜索(不基于元素对象)

        Args:
            text(str): 搜索的内容
            direction(str): "down" 或 "up"
            x(float): 横坐标滑动起始点
            y(float): 纵坐标滑动起始点
        """
        x = 0.5 if not x else x
        y = 0.5 if not y else y
        for i in range(20):
            if self.driver(textContains=text).exists:
                break
            else:
                if direction == "down":
                    self.swipe_down(fx=x, fy=y, tx=x, ty=y - 0.2, between=1)
                elif direction == "up":
                    self.swipe_up(fx=x, fy=y, tx=x, ty=y + 0.2, between=1)
                else:
                    raise NameError
        else:
            raise TimeoutError("Timeout, element not found")

    def press(self, key):
        """
        Documents:
            press key via name or key code. Supported key name includes:
            home, back, left, right, up, down, center, menu, search, enter,
            delete(or del), recent(recent apps), volume_up, volume_down,
            volume_mute, camera, power.
        """
        names = ["home", "back", "left", "right", "up", "down", "center", "menu", "search", "enter", "delete", "recent",
                 "volume_up", "volume_down", "volume_mute", "camera", "power"]
        if key in names:
            self.driver.press(key)
        else:
            raise NameError("Check the input text")

    def assert_text_exists(self, text: str, describe: str, sleep=0, timeout=10):
        """
        Asserts that the text exists on the current page

        Args：
            sleep(int): sleep time
            text(str): text
            describe(str): Assertion description information
            timeout(int): Maximum waiting time
        """
        self.sleep(sleep)

        text_exists = self.driver(text=text).exists(timeout)
        self.screenshots(describe="断言")
        logging.info("预期结果: " + describe + " 文案存在")
        if text_exists is True:
            insert_assert(describe, True)
            logging.info("实际结果: " + describe + " 文案存在")
        else:
            insert_assert(describe, False)
            logging.warning("实际结果: " + describe + " 文案不存在")

    def assert_element_exists(self, element, describe, sleep=0, timeout=10):
        """
        Asserts that the element exists on the current page

        Args：
            sleep(int): sleep time
            element(object): element object
            describe(str): Assertion description information
            timeout(int): Maximum waiting time
        """
        self.sleep(sleep)

        element_exists = element.exists(timeout)
        self.screenshots(describe="断言")
        logging.info("预期结果: " + describe + " 元素存在")
        if element_exists is True:
            insert_assert(describe, True)
            logging.info("实际结果: " + describe + " 元素存在")
        else:
            insert_assert(describe, False)
            logging.warning("实际结果: " + describe + " 元素不存在")

    def assert_text_not_exists(self, text, describe, sleep=0, timeout=10):
        """
        Asserts that the text not exists on the current page

        Args：
            sleep(int): sleep time
            text(str): text
            describe(str): Assertion description information
            timeout(int): Maximum waiting time
        """
        self.sleep(sleep)

        text_exists = self.driver(text=text).exists(timeout)
        self.screenshots(describe="断言")
        logging.info("预期结果: " + describe + " 文案不存在")
        if text_exists is True:
            insert_assert(describe, False)
            logging.warning("实际结果: " + describe + " 文案存在")
        else:
            insert_assert(describe, True)
            logging.info("实际结果: " + describe + " 文案不存在")

    def assert_element_not_exists(self, element, describe, sleep=0, timeout=10):
        """
        Asserts that the element not exists on the current page

        Args：
            sleep(int): sleep time
            element(object): element object
            describe(str): Assertion description information
        """
        self.sleep(sleep)

        element_exists = element.exists(timeout)
        self.screenshots(describe="断言")
        logging.info("预期结果: " + describe + " 元素不存在")
        if element_exists is True:
            insert_assert(describe, False)
            logging.warning("实际结果: " + describe + " 元素存在")
        else:
            insert_assert(describe, True)
            logging.info("实际结果: " + describe + " 元素不存在")

    def assert_contain_text(self, text, describe, sleep=0, timeout=10):
        """
        Asserts that two texts are not equal

        Args：
            sleep(int): sleep time
            text(str): text
            describe(str): Assertion description information
            timeout(int): Maximum waiting time
        """
        self.sleep(sleep)

        text_exists = self.driver(textContains=text).exists(timeout)
        self.screenshots(describe="断言")
        logging.info("预期结果: " + describe + " 文案存在")
        if text_exists is True:
            result = [describe, True]
            App.assert_result.append(result)
            logging.info("实际结果: " + describe + " 文案存在")
        else:
            result = [describe, False]
            App.assert_result.append(result)
            logging.warning("实际结果: " + describe + " 文案不存在")

    def assert_not_contain_text(self, text, describe, sleep=0, timeout=10):
        """
        Asserts that two texts are not equal

        Args：
            sleep(int): sleep time
            text(str): text
            describe(str): Assertion description information
            timeout(int): Maximum waiting time
        """
        self.sleep(sleep)

        text_exists = self.driver(textContains=text).exists(timeout)
        self.screenshots(describe="断言")
        logging.info("预期结果: " + describe + " 文案不存在")
        if text_exists is True:
            result = [describe, False]
            App.assert_result.append(result)
            logging.warning("实际结果: " + describe + " 文案存在")
        else:
            result = [describe, True]
            App.assert_result.append(result)
            logging.info("实际结果: " + describe + " 文案不存在")

    @staticmethod
    def assert_text_equals(text_1, text_2, describe):
        """
        Asserts that two texts are equal

        Args：
            text(list): text
        """
        logging.info("预期结果: " + text_1 + "," + text_2 + " 相等")

        if text_1 == text_2:
            result = [describe, True]
            App.assert_result.append(result)
            logging.info("实际结果: " + text_1 + "," + text_2 + " 相等")

        else:
            result = [describe, False]
            App.assert_result.append(result)
            logging.warning("实际结果: " + text_1 + "," + text_2 + " 不相等")

    @staticmethod
    def assert_text_not_equals(text_1, text_2, describe):
        """
        Asserts that two texts are not equal

        Args：
            text(list): text
        """
        logging.info("预期结果: " + text_1 + "," + text_2 + " 不相等")
        if text_1 == text_2:
            result = [describe, False]
            App.assert_result.append(result)
            logging.warning("预期结果: " + text_1 + "," + text_2 + " 相等")
        else:
            result = [describe, True]
            App.assert_result.append(result)
            logging.info("预期结果: " + text_1 + "," + text_2 + " 不相等")


class XpathElement(object):
    """
    Only CSS selectors are supported.
    https://github.com/openatx/uiautomator2/blob/master/XPATH.md

    >> from poium.u2 import Page, XpathElement
    >> class MyPage(Page):
            input = XpathElement('//android.widget.EditText')
            button = XpathElement('@com.taobao.taobao:id/fl_banner_container')
    """

    def __init__(self, xpath, index=None, timeout=10, describe=None):
        self.xpath = xpath
        self.describe = describe
        self.time_out = timeout
        if index is None:
            self.index = 0
        else:
            self.index = int(index)
        self.desc = describe

    def __get__(self, instance, owner):
        if instance is None:
            return None
        App.driver = instance.driver
        return self

    def click(self):
        """
        Click element.
        """
        App.driver.xpath(self.xpath).click()

    def set_text(self, value):
        """
        Simulates typing into the element.
        :param value: input text
        """
        App.driver.xpath(self.xpath).set_text(value)

    def get_text(self):
        """
        :return: get text from field
        """
        return App.driver.xpath(self.xpath).get_text()

    def match(self):
        """
        :return: None or matched XPathElement
        """
        return App.driver.xpath(self.xpath).match()


class Element(BaseMethod):
    """
    element class
    """

    def __init__(self, timeout=10, describe=None, **kwargs):
        self.describe = describe
        self.time_out = timeout
        if self.describe is None:
            self.describe = "NONE"
        if not kwargs:
            raise ValueError("Please specify a locator")
        self.kwargs = kwargs
        self.k, self.v = next(iter(kwargs.items()))

        if self.k not in LOCATOR_LIST:
            raise KeyError("Element positioning of type '{}' is not supported.".format(self.k))

    def __get__(self, instance, owner):
        if instance is None:
            return None
        App.driver = instance.driver
        return self

    def click(self, timeout=10, offset=None):
        """
        Click UI element.

        Args:
            timeout: seconds wait element show up
            offset: (xoff, yoff) default (0.5, 0.5) -> center
            screenshots: screenshots

        The click method does the same logic as java uiautomator does.
        1. waitForExists 2. get VisibleBounds center 3. send click event

        Raises:
            UiObjectNotFoundError
        """
        App.driver(**self.kwargs).click(timeout, offset)

    def click_exists(self, timeout=1):
        """
        The element is not clicked until it exists
        """
        return App.driver(**self.kwargs).click_exists(timeout=timeout)

    def click_more(self, sleep=.01, times=3):
        """
        连续点击
        sleep(float): 间隔时间
        times(int): 点击次数
        """
        x, y = self.center()
        for i in range(times):
            App.driver.touch.down(x, y)
            self.sleep(sleep)
            App.driver.touch.up(x, y)

    def exists(self, timeout=0):
        """
        check if the object exists in current window.
        """

        return App.driver(**self.kwargs).exists(timeout=timeout)

    def set_text(self, text):
        """
        input text
        :param text:
        """
        print("\n")
        logging.info(msg=" 键盘输入 ==> " + text)
        App.driver(**self.kwargs).set_text(text=text)

    def send_keys(self, text, clear=True):
        """
        alias of set_text
        :param text:
        :param clear:
        """
        App.driver(**self.kwargs).click()
        App.driver.send_keys(text=text, clear=clear)

    def clear_text(self):
        """
        Clear the text
        """
        App.driver(**self.kwargs).clear()

    def get_text(self):
        """
        get element text
        """
        return App.driver(**self.kwargs).get_text()

    def bounds(self):
        """
        Returns the element coordinate position
        :return: left_top_x, left_top_y, right_bottom_x, right_bottom_y
        """
        return App.driver(**self.kwargs).bounds()

    def get_position(self):
        """
        get position
        :return: x, y
        """
        h, w = App.driver.window_size()
        x, y = self.center()
        return round(x / h, 4), round(y / w, 4)

    def center(self):
        """
        Returns the center coordinates of the element
        return: center point (x, y)
        """
        return App.driver(**self.kwargs).center()

    def swipe(self, direction, times=1, steps=10):
        """
        Performs the swipe action on the UiObject.
        Swipe from center
        Args:
        steps: 1 steps is about 5ms, if set, duration will be ignore
        direction: "left", "right", "up", "down"
        times: move times
        """
        assert direction in ("left", "right", "up", "down")

        for i in range(times):
            App.driver(**self.kwargs).swipe(direction=direction, steps=steps)
        self.sleep(0.1)

    def sliding(self, h: float = None, click=False):
        """
        Args:
            h(float): The screen height 0 ~ 1
            click(bool): True click the element, False skip
        Documents:
            Determine if the element is on the current page, and if it isn't, slide down until you find it on the screen
            If present, slide the expected position
        """
        for i in range(30):
            if self.exists():
                break
            else:
                App.driver.swipe(0.5, 0.7, 0.5, 0.3)
        if h:
            if h <= 0 or h >= 1:
                raise ValueError("'h' checks the range of values, 0 < h < 1")
            else:
                scope = [h - 0.05, h + 0.05]
                for i in range(30):
                    x, y = self.get_position()
                    if scope[0] <= y <= scope[1]:
                        break
                    else:
                        if y > scope[1]:
                            App.driver.swipe(0.5, 0.5, 0.5, 0.45)
                        elif y < scope[0]:
                            App.driver.swipe(0.5, 0.45, 0.5, 0.5)

        self.click() if click else None

    @property
    def info(self):
        """
        The element information
        """
        return App.driver(**self.kwargs).info

    @property
    def count(self):
        """
        Gets the same number of elements
        """
        return App.driver(**self.kwargs).count

    def instance(self, num):
        """
        Click on the list of elements
        """
        _list = []
        self.wait()
        data = self.count
        for i in range(data):
            _list.append(i)
        element = Element(**self.kwargs, instance=_list[num])
        return element

    def wait(self, timeout=10):
        """
        Wait until UI Element exists or gone
        """
        return App.driver(**self.kwargs).wait(exists=True, timeout=timeout)

    def wait_gone(self, timeout=None):
        """ wait until ui gone
        Args:
            timeout (float): wait element gone timeout
        Returns:
            bool if element gone
        """
        return App.driver(**self.kwargs).wait_gone(timeout)



import os
import time
from time import sleep

from selenium.common.exceptions import NoAlertPresentException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.actions import interaction
from selenium.webdriver.common.actions.action_builder import ActionBuilder
from selenium.webdriver.common.actions.pointer_input import PointerInput

from poium.base import BaseMethod
from poium.base_page import BasePage
from poium.common.keyboard import KeyEvent


class Page(BasePage, BaseMethod):
    """
    Implement the APIs with javascript,
    and selenium/appium extension APIs。
    """

    def execute_script(self, js=None, *args):
        """
        Execute JavaScript scripts.
        """
        if js is None:
            raise ValueError("Please input js script")

        return self.driver.execute_script(js, *args)

    def window_scroll(self, width=None, height=None):
        """
        JavaScript API, Only support css positioning
        Setting width and height of window scroll bar.
        """
        if width is None:
            width = "0"
        if height is None:
            height = "0"
        js = f"window.scrollTo({width},{height});"
        self.execute_script(js)

    @property
    def get_title(self):
        """
        JavaScript API
        Get page title.
        """
        js = 'return document.title;'
        return self.execute_script(js)

    @property
    def get_url(self):
        """
        JavaScript API
        Get page URL.
        """
        js = "return document.URL;"
        return self.execute_script(js)

    def set_window_size(self, width=None, height=None):
        """
        selenium API
        Sets the width and height of the current window.

        :Args:
         - width: the width in pixels to set the window to
         - height: the height in pixels to set the window to

        :Usage:
            driver.set_window_size(800,600)
        """
        if width is None and height is None:
            self.driver.maximize_window()
        else:
            self.driver.set_window_size(self, width, height)

    def switch_to_parent_frame(self):
        """
        selenium API
        Switches focus to the parent context.
        Corresponding relationship with switch_to_frame () method.
        """
        self.driver.switch_to.parent_frame()

    def switch_to_window(self, index: int) -> None:
        """
        selenium API
        Switches focus to the specified window.

        :Args:
         - window: window index. 1 represents a newly opened window (0 is the first one)

        :Usage:
            self.switch_to_window(1)
        """
        all_handles = self.driver.window_handles
        self.driver.switch_to.window(all_handles[index])

    def screenshots(self, path=None, filename=None):
        """
        selenium API
        Saves a screenshots of the current window to a PNG image file
        :param path: The path to save the file
        :param filename: The file name
        """
        if path is None:
            path = os.getcwd()
        if filename is None:
            filename = str(time.time()).split(".")[0] + ".png"
        file_path = os.path.join(path, filename)
        self.driver.save_screenshot(file_path)

    def get_cookies(self):
        """
        Returns a set of dictionaries, corresponding to cookies visible in the current session.
        """
        return self.driver.get_cookies()

    def get_cookie(self, name):
        """
        Returns information of cookie with ``name`` as an object.
        """
        return self.driver.get_cookie(name)

    def add_cookie(self, cookie_dict):
        """
        Adds a cookie to your current session.
        Usage:
            add_cookie({'name' : 'foo', 'value' : 'bar'})
        """
        if isinstance(cookie_dict, dict):
            self.driver.add_cookie(cookie_dict)
        else:
            raise TypeError("Wrong cookie type.")

    def add_cookies(self, cookie_list):
        """
        Adds a cookie to your current session.
        Usage:
            cookie_list = [
                {'name' : 'foo', 'value' : 'bar'},
                {'name' : 'foo', 'value' : 'bar'}
            ]
            add_cookie(cookie_list)
        """
        if isinstance(cookie_list, list):
            for cookie in cookie_list:
                self.add_cookie(cookie)
        else:
            raise TypeError("Wrong cookie type.")

    def delete_cookie(self, name):
        """
        Deletes a single cookie with the given name.
        """
        self.driver.delete_cookie(name)

    def delete_all_cookies(self):
        """
        Delete all cookies in the scope of the session.
        Usage:
            self.delete_all_cookies()
        """
        self.driver.delete_all_cookies()

    def switch_to_app(self):
        """
        appium API
        Switch to native app.
        """
        current_context = self.driver.current_context
        if current_context != "NATIVE_APP":
            self.driver.switch_to.context('NATIVE_APP')

    def switch_to_web(self, context=None):
        """
        appium API
        Switch to web view.
        """
        current_context = self.driver.current_context
        if context is not None:
            self.driver.switch_to.context(context)
        elif "WEBVIEW" in current_context:
            return
        else:
            all_context = self.driver.contexts
            for context in all_context:
                if "WEBVIEW" in context:
                    self.driver.switch_to.context(context)
                    break
            else:
                raise NameError("No WebView found.")

    def switch_to_flutter(self):
        """
        appium API
         Switch to flutter app.
        """
        current_context = self.driver.current_context
        if current_context != "FLUTTER":
            self.driver.switch_to.context('FLUTTER')

    def key_text(self, text):
        """
        appium API
        Support input text, Chinese is not supported
        Usage:
            self.set_text("hello")
        """
        ke = KeyEvent(self.driver)
        ke.input(text)

    def key_text_capital(self, text):
        """
        appium API
        Support input capital text, Chinese is not supported
        Usage:
            self.key_text_capital("HELLO")
        """
        ke = KeyEvent(self.driver)
        ke.input_capital(text)

    def accept_alert(self):
        """
        selenium API
        Accept warning box.
        """
        self.driver.switch_to.alert.accept()

    def dismiss_alert(self):
        """
        selenium API
        Dismisses the alert available.
        """
        self.driver.switch_to.alert.dismiss()

    def alert_is_display(self):
        """
        selenium API
        Determines if alert is displayed
        """
        try:
            self.driver.switch_to.alert
        except NoAlertPresentException:
            return False
        else:
            return True

    @property
    def get_alert_text(self):
        """
        selenium API
        Get warning box prompt information.
        """
        return self.driver.switch_to.alert.text

    def move_by_offset(self, x, y, click=False):
        """
        selenium API
        Moving the mouse to an offset from current mouse position.

        :Args:
         - x: X offset to move to, as a positive or negative integer.
         - y: Y offset to move to, as a positive or negative integer.
        """
        if click is True:
            ActionChains(self.driver).move_by_offset(x, y).click().perform()
            return

        ActionChains(self.driver).move_by_offset(x, y).perform()

    def release(self):
        """
        selenium API
        Releasing a held mouse button on an element.
        """
        ActionChains(self.driver).release().perform()

    def top(self, elem, x, y, count):
        """
        appium API
        Perform a tap action on the element
        """
        actions = ActionChains(self.driver)
        actions.w3c_actions = ActionBuilder(self.driver, mouse=PointerInput(interaction.POINTER_TOUCH, "touch"))
        actions.w3c_actions.pointer_action.move_to_location(x, y)
        actions.w3c_actions.pointer_action.pointer_down()
        actions.w3c_actions.pointer_action.pause(0.1)
        actions.w3c_actions.pointer_action.release()
        actions.perform()
        sleep(2)

    def swipe(self, start_x, start_y, end_x, end_y, duration=None):
        """
        appium API
        Swipe from one point to another point, for an optional duration.
        """
        self.driver.swipe(start_x, start_y, end_x, end_y, duration)

    def back(self):
        """go back"""
        self.driver.back()

    def home(self):
        """press home"""
        self.driver.home()

    def wait(self, secs: int = 10) -> None:
        """
        Implicitly wait.All elements on the page.
        Usage:
            page.wait(10)
        """
        self.driver.implicitly_wait(secs)

    def wait_script_timeout(self, time_to_wait):
        """
        Set the amount of time that the script should wait during an
           execute_async_script call before throwing an error.
        Usage:
            page.wait_script_timeout(10)
        """
        self.driver.set_script_timeout(time_to_wait)

    def wait_page_load_timeout(self, time_to_wait):
        """
        Set the amount of time to wait for a page load to complete
           before throwing an error.
        Usage:
            page.wait_page_load_timeout(10)
        """
        self.driver.set_page_load_timeout(time_to_wait)



import sys
import time
from loguru import logger
from poium import config


class Logger:

    def __init__(self, level: str = "DEBUG", colorlog: bool = True):
        self.logger = logger
        self._colorlog = colorlog
        self._console_format = "<green>{time:YYYY-MM-DD HH:mm:ss}</> {file} <level>| {thread.name} | {level} | {message}</level>"
        self._level = level
        self.set_level(self._colorlog, self._console_format, self._level)

    def set_level(self, colorlog: bool = True, format: str = None, level: str = "DEBUG"):
        if format is None:
            format = self._console_format
        logger.remove()
        logger.add(sys.stderr, level=level, colorize=colorlog, format=format)

    def trace(self, msg: str):
        if config.printLog is True:
            now = time.strftime("%Y-%m-%d %H:%M:%S")
            print(f"{now} | TRACE | {str(msg)}")
        return self.logger.trace(msg)

    def debug(self, msg: str):
        if config.printLog is True:
            now = time.strftime("%Y-%m-%d %H:%M:%S")
            print(f"{now} | DEBUG | {str(msg)}")
        return self.logger.debug(msg)

    def info(self, msg: str):
        if config.printLog is True:
            now = time.strftime("%Y-%m-%d %H:%M:%S")
            print(f"{now} | INFO | {str(msg)}")
        return self.logger.info(msg)

    def success(self, msg: str):
        if config.printLog is True:
            now = time.strftime("%Y-%m-%d %H:%M:%S")
            print(f"{now} | SUCCESS | {str(msg)}")
        return self.logger.success(msg)

    def warning(self, msg: str):
        if config.printLog is True:
            now = time.strftime("%Y-%m-%d %H:%M:%S")
            print(f"{now} | WARNING | {str(msg)}")
        return self.logger.warning(msg)

    def error(self, msg: str):
        if config.printLog is True:
            now = time.strftime("%Y-%m-%d %H:%M:%S")
            print(f"{now} | ERROR | {str(msg)}")
        return self.logger.error(msg)

    def critical(self, msg: str):
        if config.printLog is True:
            now = time.strftime("%Y-%m-%d %H:%M:%S")
            print(f"{now} | CRITICAL | {str(msg)}")
        return self.logger.critical(msg)

    def printf(self, msg: str):
        return self.logger.success(msg)


# log level: TRACE < DEBUG < INFO < SUCCESS < WARNING < ERROR
logging = Logger(level="TRACE")



from poium.common import logging
from poium.common.exceptions import CSSFindElementError
from selenium.common.exceptions import JavascriptException


class CSSElement(object):
    """
    Only CSS selectors are supported.
    Please see help: http://www.w3school.com.cn/cssref/css_selectors.asp
    >> from page_objects import Page, CSSElements
    >> class MyPage(Page):
            input = CSSElements('.s_ipt')
            button = CSSElements('#su')
    """

    driver = None

    def __init__(self, css: str, index: int = 0, describe: str = ""):
        self.css = css
        self.index = str(index)
        self.desc = describe

    def __get__(self, instance, owner):
        if instance is None:
            return None
        global driver
        driver = instance.driver
        return self

    def _execute_javascript(self, js):
        """
        Run the javascript script
        """
        try:
            return driver.execute_script(js)
        except JavascriptException:
            raise CSSFindElementError("Element discovery failure. ", js)

    def clear(self) -> None:
        """
        JavaScript API
        Clears the text if it's a text entry element, Only support css positioning
        """
        logging.info(f"Clear input field. {self.desc}")
        js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                    elm.style.border="2px solid red";
                    elm.value = "";"""
        self._execute_javascript(js)

    def get_text(self, i: int = None) -> str:
        """
        JavaScript API
        Get element text content.
        :param i: index
        """
        if i is None:
            i = self.index
        else:
            i = str(i)
        logging.info(f"get text. {self.desc}")
        js = f"""return document.querySelectorAll("{self.css}")[{i}].textContent;"""
        return self._execute_javascript(js)

    def set_text(self, value: str) -> None:
        """
        JavaScript API
        Simulates typing into the element.
        :param value: input text
        """
        logging.info(f"set text. {self.desc}")
        js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                    elm.style.border="2px solid red";
                    elm.value = "{value}";"""
        self._execute_javascript(js)

    def click(self) -> None:
        """
        JavaScript API
        Click element.
        """
        logging.info(f"click element. {self.desc}")
        js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                   elm.style.border="2px solid red";
                   elm.click();"""
        self._execute_javascript(js)

    def click_display(self) -> None:
        """
        JavaScript API
        Click on the displayed element, otherwise skip it.
        """
        logging.info(f"Click on the displayed element. {self.desc}")
        js = 'var elm = document.querySelector("' + self.css + '");' \
             ' if(elm != null){elm.style.border="2px solid red";elm.click();}'
        self._execute_javascript(js)

    def display(self) -> None:
        """
        JavaScript API
        Display hidden elements
        """
        logging.info(f"display hidden element. {self.desc}")
        js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                    elm.style.display = "block";"""
        self._execute_javascript(js)

    def remove_attribute(self, attribute) -> None:
        """
        JavaScript API
        Remove element attribute
        :param attribute:
        """
        logging.info(f"remove element attribute. {self.desc}")
        js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                    elm.removeAttribute("{attribute}");"""
        self._execute_javascript(js)

    def set_attribute(self, attribute, value) -> None:
        """
        JavaScript API
        Setting element attribute
        :param attribute:
        :param value:
        """
        logging.info(f"setting element attribute. {self.desc}")
        js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                    elm.setAttribute("{attribute}", "{value}");
                    """
        self._execute_javascript(js)

    def clear_style(self) -> None:
        """
        JavaScript API
        Clear element styles.
        """
        logging.info(f"clear element styles. {self.desc}")
        js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                    elm.style="";"""
        self._execute_javascript(js)

    def clear_class(self) -> None:
        """
        JavaScript API
        Clear element class
        """
        logging.info(f"clear element class. {self.desc}")
        js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                     elm.removeAttribute("class");"""
        self._execute_javascript(js)

    def inner_text(self, text) -> None:
        """
        JavaScript API
        The innerText property sets the text content of the specified element, Only support css positioning
        :param text: Inserted text
        """
        logging.info(f"inner text. {self.desc}")
        js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                     elm.innerText="{text}";"""
        self._execute_javascript(js)

    def remove_child(self, child: int = 0) -> None:
        """
        JavaScript API
        Remove a node from the child node list
        :param child: child of the child node
        """
        logging.info(f"Remove a node from the child node list. {self.desc}")
        js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                    elm.removeChild(elm.childNodes[{child}]);"""
        self._execute_javascript(js)

    def click_parent(self) -> None:
        """
        JavaScript API
        Click the parent element of the element
        """
        logging.info(f"click the parent element of the element. {self.desc}")
        js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                    elm.parentElement.click();"""
        self._execute_javascript(js)

    def scroll(self, top=0, left=0) -> None:
        """
        JavaScript API
        Scroll the div element on the page
        """
        logging.info(f"scroll the div element on the page. {self.desc}")
        if top != 0:
            js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                    elm.scrollTop={top};"""
            self._execute_javascript(js)
        if left != 0:
            js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                    elm.scrollLeft={left};"""
            self._execute_javascript(js)

    def move_to(self) -> None:
        """
        JavaScript API
        Move the mouse over the element
        """
        logging.info(f"Move the mouse over the element. {self.desc}")
        js = f"""var elm = document.querySelectorAll("{self.css}")[{self.index}];
                    elm.dispatchEvent(new Event("mouseover"));"""
        self._execute_javascript(js)

    @property
    def value(self) -> str:
        """
        JavaScript API
        Gets the value of the element.
        """
        logging.info(f"get element value. {self.desc}")
        js = f"""return document.querySelectorAll("{self.css}")[{self.index}].value;"""
        return self._execute_javascript(js)



from poium.base import BaseMethod
from poium.common import logging
from poium.common.assert_des import insert_assert
from poium.config import App

LOCATOR_LIST = [
    "id",
    "name",
    "text",
    "nameContains",
    "label",
    "xpath",
    "labelContains",
    "className",
    "predicate",
    "classChain"
]


class Page(BaseMethod):
    """
    facebook-wda page class
    """

    def __init__(self, dr):
        self.driver = dr

    def native_resolution(self):
        """
        获取屏幕原始分辨率
        """
        multiple = self.driver.scale
        w, h = self.driver.window_size()
        return multiple * w, multiple * h

    def close(self):
        """
        关闭App
        """
        self.driver.close()

    def click(self, x: float = None, y: float = None, text: str = None, screenshots=App.click_screenshots):
        """
        点击坐标
        Args：
            x(float): x坐标
            y(float): y坐标
            text(str): 文本
            screenshots(bool): 当screenshots等于True， 会先截图再点击坐标；默认关闭
        """
        if not x and not y and not text:
            raise ValueError
        (x, y) = self.get_position(text=text) if text else (x, y)
        self.screenshots(x, y, describe="点击坐标, {},{}".format(x, y)) if screenshots else \
            (print("\n"), logging.info(msg=" 点击 ==> " + "点击坐标{},{}".format(x, y)))
        self.driver.click(x, y)

    def get_position(self, text=None, element=None):
        """
        获取元素或文本坐标

        Args:
            text(str): 文案
            element(object): atx原生的元素对象
        """
        w, h = self.driver.window_size()
        if text is not None:
            rect = self.driver(name=text).bounds
        elif element is not None:
            rect = element.bounds
        else:
            raise NameError
        x = rect.x + rect.width / 2
        y = rect.y + rect.height / 2
        return x / w, y / h

    def swipe(self, fx: float, fy: float, tx: float, ty: float, duration=0, screenshots=True):
        """
        滑动
        Args:
            fx(float): 起始横坐标
            fy(float): 起始纵坐标
            tx(float): 终点横坐标
            ty(float): 终点纵坐标
            duration(float): 滑动过程的时间 (seconds)
            screenshots(bool): 滑动后截图开关
        """
        self.driver.swipe(fx, fy, tx, ty, duration=duration)
        if screenshots is True:
            self.sleep(0.5)
            self.screenshots()

    def swipe_left(self, fx=0.3, fy=0.5, tx=0.7, ty=0.5, times=1, duration=0, screenshots=True):
        """
        滑向左边
        Args:
            fx(float): 起始横坐标
            fy(float): 起始纵坐标
            tx(float): 终点横坐标
            ty(float): 终点纵坐标
            times(int): 滑动的次数
            duration(float): 滑动过程的时间 (seconds)
            screenshots(bool): 滑动后截图开关
        """
        for i in range(times):
            self.swipe(fx, fy, tx, ty, duration=duration, screenshots=screenshots)

    def swipe_right(self, fx=0.7, fy=0.5, tx=0.3, ty=0.5, times=1, duration=0, screenshots=True):
        """
        滑向右边
        Args:
            fx(float): 起始横坐标
            fy(float): 起始纵坐标
            tx(float): 终点横坐标
            ty(float): 终点纵坐标
            times(int): 滑动的次数
            duration(float): 滑动过程的时间 (seconds)
            screenshots(bool): 滑动后截图开关
        """
        for i in range(times):
            self.swipe(fx, fy, tx, ty, duration=duration, screenshots=screenshots)

    def swipe_up(self, fx=0.5, fy=0.5, tx=0.5, ty=0.8, times=1, duration=0, screenshots=True):
        """
        滑向上边
        Args:
            fx(float): 起始横坐标
            fy(float): 起始纵坐标
            tx(float): 终点横坐标
            ty(float): 终点纵坐标
            times(int): 滑动的次数
            duration(float): 滑动过程的时间 (seconds)
            screenshots(bool): 滑动后截图开关
        """
        for i in range(times):
            self.swipe(fx, fy, tx, ty, duration=duration, screenshots=screenshots)

    def swipe_down(self, fx=0.5, fy=0.5, tx=0.5, ty=0.2, times=1, duration=0, screenshots=True):
        """
        滑向下边
        Args:
            fx(float): 起始横坐标
            fy(float): 起始纵坐标
            tx(float): 终点横坐标
            ty(float): 终点纵坐标
            times(int): 滑动的次数
            duration(float): 滑动过程的时间 (seconds)
            screenshots(bool): 滑动后截图开关
        """
        for i in range(times):
            self.swipe(fx, fy, tx, ty, duration=duration, screenshots=screenshots)

    def swipe_search(self, text, direction="down"):
        """
        文本搜索(不基于元素对象)

        Args:
            text(str): 搜索的内容
            direction(str): "down" 或 "up"
        """
        for i in range(20):
            if self.driver(text=text).exists and self.driver(text=text).displayed:
                break
            else:
                if direction == "down":
                    self.swipe_down()
                elif direction == "up":
                    self.swipe_up()
                else:
                    raise NameError
        else:
            raise TimeoutError("Timeout, element not found")

    def who_exists(self, element=None, text=None):
        """
        判断不同页面的多个元素或文本，看哪一个先出现，判断页面的状态
        Args：
            element(list): 元素列表，不同页面的元素对象
            text(list): 文本列表，不同页面的文本
        Return:
            element_child(object): 返回当前页面存在的元素
            text_child(text): 返回当前页面存在的文本
        """
        for i in range(10):
            if element is not None:
                if type(element) is list:
                    for element_child in element:
                        if element_child.exists() is True:
                            return element_child
                else:
                    raise TypeError("The element must be a list")
            elif text is not None:
                if type(text) is list:
                    for text_child in text:
                        if self.driver(nama=text_child).exists is True:
                            return text_child
                else:
                    raise TypeError("The text must be a list")
            else:
                raise ValueError("Must pass parameter")

            self.sleep(1)

        else:
            raise TypeError("The text or element is not exists")

    def alert(self, click=None, timeout=5) -> bool:
        for i in range(timeout):
            if "error" not in self.driver.alert.buttons():
                _list = self.driver.alert.buttons()
                text = self.driver.alert.text
                logging.info(msg="弹窗，提示⚠{text}，选项按钮{button}".format(text=text, button=_list))
                if click == "first":
                    position = self.get_position(text=_list[0])
                    self.screenshots(position[0], position[1])
                    logging.info(msg="👆 ==> {}".format(_list[0]))
                    self.driver.alert.accept()
                elif click == "second":
                    position = self.get_position(text=_list[1])
                    self.screenshots(position[0], position[1])
                    logging.info(msg="👆 ==> {}".format(_list[1]))
                    self.driver.alert.dismiss()
                else:
                    position = self.get_position(text=click)
                    self.screenshots(position[0], position[1])
                    logging.info(msg="👆 ==> {}".format(click))
                    self.driver.alert.click(click)
                return True
            else:
                self.sleep(1)
                continue
        else:
            return False

    def assert_text_exists(self, text: str, describe, sleep=0, timeout=10):
        """
        Asserts that the text exists on the current page

        Args：
            sleep(int): sleep time
            text(str): text
            describe(str): Assertion description information
            timeout(int): Maximum waiting time
        """
        self.sleep(sleep)
        logging.info("预期结果: " + describe + " 文案存在")
        for i in range(timeout):
            text_exists = self.driver(text=text).exists
            if text_exists is True:
                insert_assert(describe, True)
                logging.info("实际结果: " + describe + " 文案存在")
                break
            else:
                self.sleep(1)
                continue
        else:
            insert_assert(describe, False)
            logging.warning("实际结果: " + describe + " 文案不存在")
        self.screenshots(describe="断言")

    def assert_text_contains(self, text: str, describe, sleep=0, timeout=10):
        """
        Asserts that the text exists on the current page

        Args：
            sleep(int): sleep time
            text(str): text
            describe(str): Assertion description information
            timeout(int): Maximum waiting time
        """
        self.sleep(sleep)
        logging.info("预期结果: " + describe + " 文案存在")
        for i in range(timeout):
            text_exists = self.driver(nameContains=text).exists
            if text_exists is True:
                insert_assert(describe, True)
                logging.info("实际结果: " + describe + " 文案存在")
                break
            else:
                self.sleep(1)
                continue
        else:
            insert_assert(describe, False)
            logging.warning("实际结果: " + describe + " 文案不存在")
        self.screenshots(describe="断言")

    def assert_element_exists(self, element, describe, sleep=0, timeout=10):
        """
        Asserts that the text exists on the current page

        Args：
            sleep(int): sleep time
            element: element
            describe(str): Assertion description information
            timeout(int): Maximum waiting time
        """
        self.sleep(sleep)
        logging.info("预期结果: " + describe + " 元素存在")
        for i in range(timeout):
            element_exists = element.exists()
            if element_exists is True:
                insert_assert(describe, True)
                logging.info("实际结果: " + describe + " 元素存在")
                break
            else:
                self.sleep(1)
                continue
        else:
            insert_assert(describe, False)
            logging.warning("实际结果: " + describe + " 元素不存在")
        self.screenshots(describe="断言")

    def assert_text_not_exists(self, text: str, describe, sleep=0, timeout=10):
        """
        Asserts that the text exists on the current page

        Args：
            sleep(int): sleep time
            text(str): text
            describe(str): Assertion description information
            timeout(int): Maximum waiting time
        """
        self.sleep(sleep)
        logging.info("预期结果: " + describe + " 文案不存在")
        for i in range(timeout):
            text_exists = self.driver(text=text).exists
            if text_exists is True:
                insert_assert(describe, False)
                logging.warning("实际结果: " + describe + " 文案存在")
                break
            else:
                self.sleep(1)
                continue
        else:
            insert_assert(describe, True)
            logging.info("实际结果: " + describe + " 文案不存在")
        self.screenshots(describe="断言")

    def assert_element_not_exists(self, element, describe, sleep=0, timeout=10):
        """
        Asserts that the text exists on the current page

        Args：
            sleep(int): sleep time
            element: element
            describe(str): Assertion description information
            timeout(int): Maximum waiting time
        """
        self.sleep(sleep)
        logging.info("预期结果: " + describe + " 元素不存在")
        for i in range(timeout):
            element_exists = element.exists()
            if element_exists is True:
                insert_assert(describe, False)
                logging.warning("实际结果: " + describe + " 元素存在")
                break
            else:
                self.sleep(1)
                continue
        else:
            insert_assert(describe, True)
            logging.info("实际结果: " + describe + " 元素不存在")
        self.screenshots(describe="断言")

    @staticmethod
    def assert_text_equals(text_1, text_2, describe):
        """
        Asserts that two texts are equal

        Args：
            text(list): text
        """
        logging.info("预期结果: " + text_1 + "," + text_2 + " 相等")

        if text_1 == text_2:
            result = [describe, True]
            App.assert_result.append(result)
            logging.info("预期结果: " + text_1 + "," + text_2 + " 相等")
        else:
            result = [describe, False]
            App.assert_result.append(result)
            logging.warning("预期结果: " + text_1 + "," + text_2 + " 不相等")

    @staticmethod
    def assert_text_not_equals(text_1, text_2, describe):
        """
        Asserts that two texts are not equal

        Args：
            text(list): text
        """
        logging.info("预期结果: " + text_1 + "," + text_2 + " 不相等")

        if text_1 == text_2:
            result = [describe, False]
            App.assert_result.append(result)
            logging.warning("预期结果: " + text_1 + "," + text_2 + " 相等")
        else:
            result = [describe, True]
            App.assert_result.append(result)
            logging.info("预期结果: " + text_1 + "," + text_2 + " 不相等")


class Element(object):
    """
    element class
    """

    driver = None

    def __init__(self, timeout=10, describe=None, **kwargs):
        self.time_out = timeout
        self.describe = describe
        if not kwargs:
            raise ValueError("Please specify a locator")
        self.kwargs = kwargs
        self.k, self.v = next(iter(kwargs.items()))

        if self.k not in LOCATOR_LIST:
            raise KeyError("Element positioning of type '{}' is not supported.".format(self.k))

    def __get__(self, instance, owner):
        if instance is None:
            return None

        self.driver = instance.driver
        return self

    def click(self, focus=None, beyond=None):

        """
        点击元素, 根据坐标去点击
        Args:
            focus(list): 点击元素区域的位置，默认点击元素的中心
            beyond(list): 以传的元素为基准，点击相该元素以外的其他位置
        """

        # 通过坐标点击
        w, h = self.driver.window_size()
        if self.k == "focus":
            if type(self.v) is not list:
                raise ValueError("The argument must be a list")
            elif self.v[0] > 1 or self.v[1] > 1:
                raise ValueError
            self.driver.click(self.v[0], self.v[1])
        else:
            if focus is not None:
                x, y = self.focus(focus)
            elif beyond is not None:
                xx, yy = self.get_position(percentage=False)
                x, y = xx + beyond[0] * w, yy + beyond[1] * h
            else:
                x, y = self.focus([0.5, 0.5])

            self.driver.click(x / w, y / h)

    def click_exists(self, timeout=0):
        """
        元素存在，点击元素，不存在就pass
        Args:
            timeout(int): 最大等待时间
        """

        return self.driver(**self.kwargs).click_exists(timeout)

    def wait(self, timeout=10):
        """
        等待元素出现

        Args：
            timeout(int)：等待时间
        """
        self.driver(**self.kwargs).wait(timeout=timeout)

    def get(self, timeout=10, raise_error=False):
        """
        Args:
            timeout (float): timeout for query element, unit seconds
                Default 10s
            raise_error (bool): whether to raise error if element not found

        Returns:
            Element: UI Element

        Raises:
            WDAElementNotFoundError if raise_error is True else None
        """
        self.driver(**self.kwargs).get(timeout=timeout, raise_error=raise_error)

    def wait_gone(self, timeout=10):
        """
        等待元素消失

        Args：
            timeout(int)：等待时间
        """
        self.driver(**self.kwargs).wait_gone(timeout=timeout)

    def find_elements(self, text=False):
        """
        查找元素

        Args：
            text(bool): 返回元素对应的文本内容
        """
        text_list = []
        data = self.driver(**self.kwargs).find_elements()
        logging.info("查找到匹配数量有==> {}个".format(len(data)))
        if text is True:
            for text_data in data:
                text_list.append(text_data.get_text())
            return text_list
        else:
            return data

    def instance(self, num=1):
        """
        Click on the list of elements
        """
        _list = []
        self.wait()
        data = len(self.find_elements())
        for i in range(data):
            _list.append(i)
        if self.k == "xpath":
            self.kwargs = {"xpath": self.v + "[{}]".format(_list[num] + 1)}
            element = Element(**self.kwargs)
        else:
            element = Element(**self.kwargs, index=_list[num] + 1)
        return element

    def clear_text(self):
        """
        清空输入框
        """
        self.driver(**self.kwargs).clear_text()

    def set_text(self, text):
        """
        输入文本内容
        Args:
            text(str): 输入栏输入的文本
        """
        text = str(text)
        self.clear_text()
        logging.info(msg=" 键盘输入 ==> " + text)
        self.driver(**self.kwargs).set_text(text)

    def get_text(self):
        """
        获取元素对应的文本
        """
        return self.driver(**self.kwargs).text

    def swipe(self, direction, times=1, distance=1.0):
        """
        基于元素滑动

        times(int): 滑动次数
        distance(float): 滑动距离
        """
        assert direction in ("left", "right", "up", "down")

        for i in range(times):
            self.driver(**self.kwargs).scroll(direction=direction, distance=distance)

    def focus(self, position):
        """
        定位元素区域内的坐标
        Args:
            position(list): 元素板块内的坐标
        """
        self.get()
        if type(position) is not list:
            raise NameError("The argument must be a list")
        elif position[0] > 1 or position[1] > 1:
            raise NameError("Coordinates range from 0 to 1")
        rect = self.driver(**self.kwargs).bounds
        x = rect.x + rect.width * position[0]
        y = rect.y + rect.height * position[1]
        return x, y

    def get_position(self, percentage=True):
        """
        获取元素坐标
        Args:
            percentage(bool): percentage等于True,坐标是百分比； 默认是真实坐标
        """
        self.get()
        w, h = self.driver.window_size()
        rect = self.driver(**self.kwargs).bounds
        x = rect.x + rect.width / 2
        y = rect.y + rect.height / 2
        if percentage is True:
            return round(x / w, 6), round(y / h, 6)
        elif percentage is False:
            return x, y

    def exists(self):
        """
        判断元素是否存在
        """
        if "index" in self.kwargs:
            return True if len(self.find_elements()) > 0 else False
        else:
            return True if self.driver(**self.kwargs).exists and self.driver(**self.kwargs).displayed else False

    def scroll(self, direction='visible', distance=1.0):
        """
        滚动定位到对应的元素
        Args:
            direction (str): one of "visible", "up", "down", "left", "right"
            distance (float): swipe distance, only works when direction is not "visible"

        Raises:
            ValueError

        distance=1.0 means, element (width or height) multiply 1.0
        """
        self.driver(**self.kwargs).scroll(direction=direction, distance=distance)

    def scroll_search(self, click=False, direction="down"):
        """
        滚动定位到对应的元素

        Args:
            click(bool): 定位到元素后，是否点击
            direction(str): 滑动的方向，只能是'down' 或 'or'
        """
        for i in range(20):
            if self.exists() is True:
                break
            else:
                if direction == "down":
                    self.driver.swipe(0.5, 0.5, 0.5, 0.4)
                elif direction == "up":
                    self.driver.swipe(0.5, 0.5, 0.5, 0.6)
                else:
                    raise ValueError("The direction parameter can only be 'down' or 'up'")
        if click is True:
            self.click()

    def tap_hold(self, duration=1.0):
        """
        长按

        Args:
            duration (float): seconds of hold time
        """
        self.driver(**self.kwargs).tap_hold(duration=duration)

    def sliding(self, height=0.5, click=False, direction="down"):
        """
        将元素滑动到想要的位置
        Args:
            height(float): 预期将元素滑动到的位置， 位置的范围是 0 ~ 1， 默认是中间
            click(bool): 当click等于True，把元素滑动到预期的位置后，进行点击操作； 默认不点击，只滑动到预期的位置
            direction(str): 滑动的方向，只能是'down' 或 'or'
        """
        if 0 < height < 1:
            height = height
            height_max = height + 0.05
            height_min = height - 0.05
            if direction == "down":
                self.scroll_search()
            elif direction == "up":
                self.scroll_search(direction="up")
            else:
                raise ValueError("The direction parameter can only be 'down' or 'up'")
            x, y = self.get_position()
            for i in range(20):
                if height_min <= y <= height_max:
                    break
                move_y = height - y
                if move_y > 0:
                    if move_y >= 0.26:
                        self.driver.swipe(0.5, 0.5, 0.5, 0.6)
                    elif move_y < 0.26:
                        self.driver.swipe(0.5, 0.5, 0.5, 0.52, duration=0.5)
                elif move_y < 0:
                    if move_y <= -0.26:
                        self.driver.swipe(0.5, 0.5, 0.5, 0.4)
                    elif move_y < 0.26:
                        self.driver.swipe(0.5, 0.5, 0.5, 0.48, duration=0.5)
                x, y = self.get_position()

            if click is True:
                self.click()
        else:
            raise ValueError



from poium.config import App


def insert_assert(describe, result):
    """
    Insert assertion data
        describe(str): Assertion description information
        result(bool): Assertion results
    """
    result = [describe, bool(result)]
    App.assert_result.append(result)



"""
App keyboard operation
"""

keycodes = {
    '0': 7,
    '1': 8,
    '2': 9,
    '3': 10,
    '4': 11,
    '5': 12,
    '6': 13,
    '7': 14,
    '8': 15,
    '9': 16,
    'A': 29,
    'B': 30,
    'C': 31,
    'D': 32,
    'E': 33,
    'F': 34,
    'G': 35,
    'H': 36,
    'I': 37,
    'J': 38,
    'K': 39,
    'L': 40,
    'M': 41,
    'N': 42,
    'O': 43,
    'P': 44,
    'Q': 45,
    'R': 46,
    'S': 47,
    'T': 48,
    'U': 49,
    'V': 50,
    'W': 51,
    'X': 52,
    'Y': 53,
    'Z': 54,
    ' ': 62,
    '*': 17,
    '#': 18,
    ',': 55,
    '`': 68,
    '-': 69,
    '[': 71,
    ']': 72,
    '\\': 73,
    ';': 74,
    '/': 76,
    '@': 77,
    '=': 161,
    '.': 158,
    '+': 157,
    'NUM_LOCK': 143,   # lower case
    'CAPS_LOCK': 115,  # upper case
    'HOME': 4,    # home
    'BACK': 3,    # back
}


class KeyEvent:
    """
    KeyEvent:
    https://developer.android.com/reference/android/view/KeyEvent
    """

    def __init__(self, driver):
        self.driver = driver

    def input(self, text=""):
        """
        input text
        """
        if text == "":
            return

        for s in text:
            keycode = keycodes.get(s.upper(), 0)
            if keycode == 0:
                raise KeyError(f"The '{s}' character is not supported")
            self.driver.keyevent(keycode)

    def input_capital(self, text=""):
        """
        input capital text
        """
        if text == "":
            return

        for s in text:
            keycode = keycodes.get(s.upper(), 0)
            if keycode == 0:
                raise KeyError(f"The '{s}' character is not supported")
            if s.isupper():
                self.driver.press_keycode(keycode, 64, 59)
            else:
                self.driver.keyevent(keycode)



import os

current_path = os.path.abspath(__file__)
BASE_DIR = os.path.abspath(os.path.dirname(current_path) + os.path.sep)

# 开启print打印
printLog = False


class Browser:
    """
    web config
    """
    # Default browser driver (abandon)
    driver = None

    # Default playwright page driver (abandon)
    page = None

    # Adds a border to the action element of the operation
    show = True


class App:
    """
    App config
    """

    driver = None

    # 是否通过usb链接设备
    connect_usb = True

    # 设备ID，deviceID，如果为空，则自动获取连接本地pc的第一个可用设备
    device_id = ""

    # 打开APP的包名
    apk_name = ""

    # 默认点击截图开关
    click_screenshots = False

    # 报告保存路径
    report_path = "."

    # 报告截图路径
    report_snapshot = "."

    # 断言结果
    assert_result = []

    # 打开APP等待启动时间
    app_wait = 10

    # 设置元素查找等待时间
    implicitly_wait = 10

    # 设置首页加载完成等待时间
    loading_finish = 10



"""
author：bugmaster

"""
import platform
from time import sleep

from appium.webdriver.common.appiumby import AppiumBy
from func_timeout import func_set_timeout
from func_timeout.exceptions import FunctionTimedOut
from selenium.common.exceptions import StaleElementReferenceException
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.select import Select

from poium import config
from poium.common import logging
from poium.common.exceptions import PageElementError, FindElementTypesError, DriverNoneException
from poium.common.selector import selection_checker
from poium.config import Browser

# Map PageElement constructor arguments to webdriver locator enums
LOCATOR_LIST = {
    # selenium
    'css': By.CSS_SELECTOR,
    'id_': By.ID,
    'name': By.NAME,
    'xpath': By.XPATH,
    'link_text': By.LINK_TEXT,
    'partial_link_text': By.PARTIAL_LINK_TEXT,
    'tag': By.TAG_NAME,
    'class_name': By.CLASS_NAME,
    # appium
    'ios_uiautomation': AppiumBy.IOS_UIAUTOMATION,
    'ios_predicate': AppiumBy.IOS_PREDICATE,
    'ios_class_chain': AppiumBy.IOS_CLASS_CHAIN,
    'android_uiautomator': AppiumBy.ANDROID_UIAUTOMATOR,
    'android_viewtag': AppiumBy.ANDROID_VIEWTAG,
    'android_data_matcher': AppiumBy.ANDROID_DATA_MATCHER,
    'android_view_matcher': AppiumBy.ANDROID_VIEW_MATCHER,
    'accessibility_id': AppiumBy.ACCESSIBILITY_ID,
    'image': AppiumBy.IMAGE,
    'custom': AppiumBy.CUSTOM,
}

BY_LIST = [
    # selenium
    By.CSS_SELECTOR,
    By.ID,
    By.NAME,
    By.XPATH,
    By.LINK_TEXT,
    By.PARTIAL_LINK_TEXT,
    By.TAG_NAME,
    By.CLASS_NAME,
    # appium
    AppiumBy.IOS_UIAUTOMATION,
    AppiumBy.IOS_PREDICATE,
    AppiumBy.IOS_CLASS_CHAIN,
    AppiumBy.ANDROID_UIAUTOMATOR,
    AppiumBy.ANDROID_VIEWTAG,
    AppiumBy.ANDROID_DATA_MATCHER,
    AppiumBy.ANDROID_VIEW_MATCHER,
    AppiumBy.ACCESSIBILITY_ID,
    AppiumBy.IMAGE,
    AppiumBy.CUSTOM
]


class BasePage:
    """
    Page Object pattern base class.
    """

    def __init__(self, driver=None, url: str = None, print_log: bool = False):
        """
        :param driver: `selenium.webdriver.WebDriver` Selenium webdriver instance
        :param url: `str`
        :param print_log: `bool` Need to be turned on when used with the seldom framework
        """
        self.driver = None
        if driver is not None:
            self.driver = driver
        else:
            try:
                # support seldom driver
                from seldom import Seldom
                if Seldom.driver is not None:
                    self.driver = Seldom.driver
                    print_log = True
            except ImportError:
                ...

        if self.driver is None:
            raise DriverNoneException("driver is None, Please set selenium/appium driver.")
        self.root_uri = url if url else getattr(self.driver, 'url', None)
        config.printLog = print_log

    def open(self, uri: str) -> None:
        """
        open uri
        :param uri: URI to GET, based off of the root_uri attribute.
        :return:
        """
        root_uri = self.root_uri or ''
        self.driver.get(root_uri + uri)
        self.driver.implicitly_wait(5)


class Element(object):
    """
    Returns an element object
    """

    def __init__(self, selector: str = None, timeout: int = 5, describe: str = "", index: int = 0, **kwargs):
        self.selector = selector
        self.times = timeout
        self.index = index
        self.desc = describe
        self.exist = False

        if selector is not None:
            self.by, self.value = selection_checker(selector)
        else:
            if not kwargs:
                raise ValueError("Please specify a locator")
            if len(kwargs) > 1:
                raise ValueError("Please specify only one locator")
            self.kwargs = kwargs
            self.by, self.value = next(iter(kwargs.items()))

            self.by = LOCATOR_LIST.get(self.by, None)
            if self.by is None:
                raise FindElementTypesError("Element positioning of type '{}' is not supported.".format(self.by))

    def __get__(self, instance, owner):
        if instance is None:
            return None

        self.driver = instance.driver
        return self

    def __set__(self, instance, value):
        self.__get__(instance, instance.__class__)
        self.send_keys(value)

    @func_set_timeout(1)
    def find_elements_timeout(self, by: str, value: str):
        """
        set find element timeout.
        :param by:
        :param value:
        :return:
        """
        return self.driver.find_elements(by, value)

    def find(self, by: str, value: str) -> list:
        """
        Find if the element exists.
        """
        for i in range(self.times):
            try:
                elems = self.find_elements_timeout(by, value)
                break
            except FunctionTimedOut:
                sleep(1)
        else:
            elems = []

        if len(elems) == 1:
            logging.info(f"🔍 Find element: {by}={value}. {self.desc}")
        elif len(elems) > 1:
            logging.info(f"❓ Find {len(elems)} elements through: {by}={value}. {self.desc}")
        else:
            logging.warning(f"❌ Find 0 elements through: {by}={value}. {self.desc}")

        return elems

    def get_element_object(self, by: str = None, value: str = None):
        """
        Judge element positioning way, and returns the element.
        """
        if by is None or value is None:
            by = self.by
            value = self.value

        if by not in BY_LIST:
            raise FindElementTypesError("Please enter the correct targeting elements")

        elems = self.find(by, value)
        if len(elems) == 0:
            self.exist = False
            return None
        else:
            self.exist = True
            elem = elems[self.index]

        if Browser.show is True:
            try:
                style_red = 'arguments[0].style.border="2px solid #FF0000"'
                style_blue = 'arguments[0].style.border="2px solid #00FF00"'
                style_null = 'arguments[0].style.border=""'

                for _ in range(2):
                    self.driver.execute_script(style_red, elem)
                    sleep(0.1)
                    self.driver.execute_script(style_blue, elem)
                    sleep(0.1)
                self.driver.execute_script(style_blue, elem)
                sleep(0.5)
                self.driver.execute_script(style_null, elem)
            except WebDriverException:
                pass

        return elem

    def is_exist(self) -> bool:
        """element is existed """
        self.get_element_object()
        return self.exist

    def clear(self) -> None:
        """Clears the text if it's a text entry element."""
        logging.info("✅ clear.")
        elem = self.get_element_object()
        elem.clear()

    def send_keys(self, value, clear=False, click=False) -> None:
        """
        Simulates typing into the element.
        If clear_before is True, it will clear the content before typing.
        """
        elem = self.get_element_object()
        if click is True:
            elem.click()
            sleep(0.5)
            logging.info(f"✅ click().")
        if clear is True:
            elem.clear()
            sleep(0.5)
            logging.info(f"✅ clear().")
        elem.send_keys(value)
        logging.info(f"✅ send_keys('{value}').")

    def click(self) -> None:
        """
        Clicks the element.
        """
        elem = self.get_element_object()
        elem.click()
        logging.info(f"✅ click().")

    def submit(self):
        """
        Submits a form.
        """
        elem = self.get_element_object()
        elem.submit()
        logging.info(f"✅ submit().")

    @property
    def tag_name(self) -> str:
        """This element's ``tagName`` property."""
        elem = self.get_element_object()
        tag_name = elem.tag_name
        logging.info(f"✅ tag_name: {tag_name}.")
        return tag_name

    @property
    def text(self) -> str:
        """The text of the element."""
        elem = self.get_element_object()
        text = elem.text
        logging.info(f"✅ text: {text}.")
        return text

    @property
    def size(self) -> dict:
        """The size of the element."""
        elem = self.get_element_object()
        size = elem.size
        logging.info(f"✅ size: {size}.")
        return size

    @property
    def location(self) -> dict:
        """The location of the element in the renderable canvas."""
        elem = self.get_element_object()
        loc = elem.location
        logging.info(f"✅ location: {loc}.")
        return loc

    @property
    def rect(self) -> dict:
        """A dictionary with the size and location of the element."""
        elem = self.get_element_object()
        rect = elem.rect
        logging.info(f"✅ rect: {rect}.")
        return rect

    @property
    def aria_role(self) -> str:
        """Returns the ARIA role of the current web element."""
        elem = self.get_element_object()
        aria_role = elem.aria_role
        logging.info(f"✅ aria_role: {aria_role}.")
        return aria_role

    @property
    def accessible_name(self) -> str:
        """Returns the ARIA Level of the current webelement."""
        elem = self.get_element_object()
        accessible_name = elem.accessible_name
        logging.info(f"✅ accessible_name: {accessible_name}.")
        return accessible_name

    def screenshot(self, filename: str = None):
        """
        element screenshot.
        :param filename:
        :return:
        """
        elem = self.get_element_object()
        if filename is not None:
            result = elem.screenshot(filename)
            logging.info(f"✅ screenshot save to: {filename}.")
            return result

        b64 = elem.screenshot_as_png
        logging.info(f"✅ screenshot save base64.")
        return b64

    def value_of_css_property(self, property_name):
        """
        The value of a CSS property
        :param property_name:
        """
        elem = self.get_element_object()
        property_value = elem.value_of_css_property(property_name)
        logging.info(f"✅ value_of_css_property('{property_name}') -> {property_value}.")
        return property_value

    def get_property(self, name) -> str:
        """
        Gets the given property of the element.
        """
        elem = self.get_element_object()
        value = elem.get_property(name)
        logging.info(f"✅ get_property('{name}') -> {value}.")
        return value

    def get_dom_attribute(self, name) -> str:
        """Gets the given attribute of the element."""
        elem = self.get_element_object()
        value = elem.get_dom_attribute(name)
        logging.info(f"✅ get_dom_attribute('{name}') -> {value}.")
        return value

    def get_attribute(self, name) -> str:
        """
        Gets the given attribute or property of the element.
        """
        elem = self.get_element_object()
        value = elem.get_attribute(name)
        logging.info(f"✅ get_attribute('{name}') -> {value}.")
        return value

    def is_displayed(self) -> bool:
        """Whether the element is visible to a user."""
        elem = self.get_element_object()
        display = elem.is_displayed()
        logging.info(f"✅ is_displayed() -> {display}.")
        return display

    def is_selected(self) -> bool:
        """
        Returns whether the element is selected.

        Can be used to check if a checkbox or radio button is selected.
        """
        elem = self.get_element_object()
        select = elem.is_selected()
        logging.info(f"✅ is_selected() -> {select}.")
        return select

    def is_enabled(self) -> bool:
        """Returns whether the element is enabled."""
        elem = self.get_element_object()
        enable = elem.is_enabled()
        logging.info(f"✅ is_enabled() -> {enable}.")
        return enable

    def switch_to_frame(self) -> None:
        """
        selenium API
        Switches focus to the specified frame
        """
        elem = self.get_element_object()
        self.driver.switch_to.frame(elem)
        logging.info(f"✅ switch_to_frame().")

    def frame(self) -> None:
        """
        selenium API
        Switches focus to the specified frame
        """
        return self.switch_to_frame()

    def move_to_element(self) -> None:
        """
        selenium API
        Moving the mouse to the middle of an element
        """
        elem = self.get_element_object()
        ActionChains(self.driver).move_to_element(elem).perform()
        logging.info(f"✅ move_to_element().")

    def click_and_hold(self) -> None:
        """
        selenium API
        Holds down the left mouse button on an element.
        """
        elem = self.get_element_object()
        ActionChains(self.driver).click_and_hold(elem).perform()
        logging.info(f"✅ click_and_hold().")

    def double_click(self) -> None:
        """
        selenium API
        Holds down the left mouse button on an element.
        """
        elem = self.get_element_object()
        ActionChains(self.driver).double_click(elem).perform()
        logging.info(f"✅ double_click().")

    def context_click(self) -> None:
        """
        selenium API
        Performs a context-click (right click) on an element.
        """
        elem = self.get_element_object()
        ActionChains(self.driver).context_click(elem).perform()
        logging.info(f"✅ context_click().")

    def drag_and_drop_by_offset(self, x: int, y: int) -> None:
        """
        selenium API
        Holds down the left mouse button on the source element,
           then moves to the target offset and releases the mouse button.
        :param x: X offset to move to.
        :param y: Y offset to move to.
        """
        elem = self.get_element_object()
        ActionChains(self.driver).drag_and_drop_by_offset(elem, xoffset=x, yoffset=y).perform()
        logging.info(f"✅ drag_and_drop_by_offset('{x}', '{y}').")

    def refresh_element(self, timeout: int = 10) -> None:
        """
        selenium API
        Refreshes the current page, retrieve elements.
        """
        elem = self.get_element_object()
        for i in range(timeout):
            if elem is not None:
                try:
                    elem
                except StaleElementReferenceException:
                    self.driver.refresh()
                else:
                    break
            else:
                sleep(1)
        else:
            raise TimeoutError("stale element reference: element is not attached to the page document.")

    def select_by_value(self, value: str) -> None:
        """
        selenium API
        Select all options that have a value matching the argument. That is, when given "foo" this
           would select an option like:

           <option value="foo">Bar</option>

           :Args:
            - value - The value to match against

           throws NoSuchElementException If there is no option with specisied value in SELECT
        """
        select_elem = self.get_element_object()
        Select(select_elem).select_by_value(value)
        logging.info(f"✅ select_by_value('{value}').")

    def select_by_index(self, index: int) -> None:
        """
        selenium API
        Select the option at the given index. This is done by examing the "index" attribute of an
           element, and not merely by counting.

           :Args:
            - index - The option at this index will be selected

           throws NoSuchElementException If there is no option with specisied index in SELECT
        """
        select_elem = self.get_element_object()
        Select(select_elem).select_by_index(index)
        logging.info(f"✅ select_by_index('{index}').")

    def select_by_visible_text(self, text: str) -> None:
        """
        selenium API
        Select all options that display text matching the argument. That is, when given "Bar" this
           would select an option like:

            <option value="foo">Bar</option>

           :Args:
            - text - The visible text to match against

            throws NoSuchElementException If there is no option with specisied text in SELECT
        """
        select_elem = self.get_element_object()
        Select(select_elem).select_by_visible_text(text)
        logging.info(f"✅ select_by_visible_text('{text}').")

    def set_text(self, keys):
        """
        appium API
        Sends text to the element.
        """
        elem = self.get_element_object()
        elem.set_text(keys)
        logging.info(f"✅ set_text('{keys}').")
        return self

    @property
    def location_in_view(self):
        """
        appium API
        Gets the location of an element relative to the view.
        Returns:
            dict: The location of an element relative to the view
        """
        elem = self.get_element_object()
        location = elem.location_in_view
        logging.info(f"✅ location_in_view -> {location}.")
        return location

    def set_value(self, value: str):
        """
        appium API
        Set the value on this element in the application
        """
        elem = self.get_element_object()
        elem.set_value(value)
        logging.info(f"✅ set_value('{value}').")
        return self

    def input(self, text="") -> None:
        """
        input.
        :param text:
        :return:
        """
        elem = self.get_element_object()
        elem.send_keys(text)
        logging.info(f"🎹 input('{text}').")

    def enter(self) -> None:
        """
        Enter key
        :return:
        """
        elem = self.get_element_object()
        elem.send_keys(Keys.ENTER)
        logging.info(f"🎹 enter.")

    def select_all(self) -> None:
        """
        select all.
        :return:
        """
        elem = self.get_element_object()
        if platform.system().lower() == "darwin":
            elem.send_keys(Keys.COMMAND, "a")
        else:
            elem.send_keys(Keys.CONTROL, "a")
        logging.info(f"🎹 control + a.")

    def cut(self) -> None:
        """
        cut.
        :return:
        """
        elem = self.get_element_object()
        if platform.system().lower() == "darwin":
            elem.send_keys(Keys.COMMAND, "x")
        else:
            elem.send_keys(Keys.CONTROL, "x")
        logging.info(f"🎹 control + x.")

    def copy(self) -> None:
        """
        copy
        :return:
        """
        elem = self.get_element_object()
        if platform.system().lower() == "darwin":
            elem.send_keys(Keys.COMMAND, "c")
        else:
            elem.send_keys(Keys.CONTROL, "c")
        logging.info(f"🎹 control + c.")

    def paste(self) -> None:
        """
        paste.
        :return:
        """
        elem = self.get_element_object()
        if platform.system().lower() == "darwin":
            elem.send_keys(Keys.COMMAND, "v")
        else:
            elem.send_keys(Keys.CONTROL, "v")
        logging.info(f"🎹 control + v.")

    def backspace(self) -> None:
        """
        Backspace key.
        :return:
        """
        elem = self.get_element_object()
        elem.send_keys(Keys.BACKSPACE)
        logging.info(f"🎹 backspace.")

    def delete(self) -> None:
        """
        Delete key.
        :return:
        """
        elem = self.get_element_object()
        elem.send_keys(Keys.DELETE)
        logging.info(f"🎹 delete.")

    def tab(self) -> None:
        """
        Tab key.
        :return:
        """
        elem = self.get_element_object()
        elem.send_keys(Keys.TAB)
        logging.info(f"🎹 tab.")

    def space(self) -> None:
        """
        Space key.
        :return:
        """
        elem = self.get_element_object()
        elem.send_keys(Keys.SPACE)
        logging.info(f"🎹 space.")


class Elements(object):
    """
    Returns a set of element objects
    """

    def __init__(self, selector: str = None, context: bool = False, describe: str = "", timeout: int = 5, **kwargs):
        self.desc = describe
        self.times = timeout
        if selector is not None:
            self.by, self.value = selection_checker(selector)
        else:
            if not kwargs:
                raise ValueError("Please specify a locator")
            if len(kwargs) > 1:
                raise ValueError("Please specify only one locator")
            by, self.value = next(iter(kwargs.items()))

            self.by = LOCATOR_LIST.get(by, None)
            if self.by is None:
                raise FindElementTypesError("Element positioning of type '{}' is not supported.".format(self.by))

        self.has_context = bool(context)

    def find(self, context):
        """
        find element.
        :param context:
        :return:
        """
        for i in range(self.times):
            elems = context.find_elements(self.by, self.value)
            if len(elems) > 0:
                break
            else:
                sleep(1)
        else:
            elems = []

        logging.info(f"✨ Find {len(elems)} elements through: {self.by}={self.value}. {self.desc}")
        return elems

    def __get__(self, instance, owner, context=None):
        if not instance:
            return None

        if not context and self.has_context:
            return lambda ctx: self.__get__(instance, owner, context=ctx)

        if not context:
            context = instance.driver

        return self.find(context)

    def __set__(self, instance, value):
        if self.has_context:
            raise PageElementError("Sorry, the set descriptor doesn't support elements with context.")
        elems = self.__get__(instance, instance.__class__)
        if not elems:
            raise PageElementError("Can't set value, no elements found")
        [elem.send_keys(value) for elem in elems]



import time


class BaseMethod:
    """
    Define basic generic methods
    """

    @staticmethod
    def sleep(sec) -> None:
        """
        sleep
        :param sec:
        :return:
        """
        time.sleep(sec)



try:
    import wda
except ImportError:
    raise ImportError("Please install 'facebook-wda' library")

from poium.common import logging
from poium.config import App

wda.DEBUG = False  # default False
wda.HTTP_TIMEOUT = 180.0  # default 60.0 seconds


def connect():
    """
    链接iOS设备
    """
    driver = wda.Client(App.device_id)
    logging.info("📱📱📱 info ===> {}!".format(driver.status()))

    return driver


def start_app(apk=None):
    """
    start app.
    :param apk:
    :return:
    """
    if apk is None:
        apk = App.apk_name
    driver = connect()
    sess = driver.session(apk)

    return sess
